from http import HTTPStatus
from json import JSONDecodeError

import pytest

from arkindex.exceptions import (
    ArkindexClientError,
    ArkindexConflictError,
    ArkindexServerError,
    ArkindexValidationError,
    ErrorResponse,
)


def test_request_no_more_retries(responses, api_client):
    # First call is an error
    responses.add(
        responses.POST,
        "https://dummy.test/api/v1/user/login/",
        status=502,
        body="Not JSON!",
    )

    api_client.request.retry.wait = 0

    # There should be 5 retries
    max_retries = 5

    with pytest.raises(ErrorResponse) as exc_info:
        api_client.login("user@user.user", "Pa$$w0rd")

    assert exc_info.value.title == "502 Bad Gateway"
    assert exc_info.value.status_code == 502
    # The raw contents of the body are output in the error
    assert exc_info.value.content == "Not JSON!"

    assert len(responses.calls) == max_retries
    assert [call.request.url for call in responses.calls] == [
        "https://dummy.test/api/v1/user/login/",
    ] * max_retries


def test_non_5xx_parse_error(responses, api_client):
    """
    For any HTTP response that is not 5xx, a parse error should be re-raised
    """
    responses.add(
        responses.POST,
        "https://dummy.test/api/v1/user/login/",
        status=402,
        # Explicitly remove the Content-Type header
        content_type=None,
        body="Not JSON!",
    )

    with pytest.raises(JSONDecodeError):
        api_client.login("user@user.user", "Pa$$w0rd")

    assert len(responses.calls) == 1
    assert responses.calls[0].request.url == "https://dummy.test/api/v1/user/login/"


@pytest.mark.parametrize(
    ("status_code", "data", "expected_class"),
    [
        (
            400,
            {
                "type": "client_error",
                "errors": [{"attr": None, "code": "invalid", "detail": "Wrong!"}],
            },
            ArkindexClientError,
        ),
        # A ConflictError requires both 409 and the `existing_id`, otherwise it falls back to ClientError
        (
            409,
            {
                "type": "client_error",
                "errors": [{"attr": ["a", 1], "code": "invalid", "detail": "Wrong!"}],
            },
            ArkindexClientError,
        ),
        (
            400,
            {
                "type": "client_error",
                "errors": [{"attr": ["a", 1], "code": "invalid", "detail": "Wrong!"}],
                "existing_id": "abc",
            },
            ArkindexClientError,
        ),
        (
            409,
            {
                "type": "client_error",
                "errors": [{"attr": ["a", 1], "code": "invalid", "detail": "Wrong!"}],
                "existing_id": None,
            },
            ArkindexConflictError,
        ),
        (
            409,
            {
                "type": "client_error",
                "errors": [{"attr": ["a", 1], "code": "invalid", "detail": "Wrong!"}],
                "existing_id": "abc",
            },
            ArkindexConflictError,
        ),
        (
            400,
            {
                "type": "validation_error",
                "errors": [{"attr": ["a", 1], "code": "invalid", "detail": "Wrong!"}],
            },
            ArkindexValidationError,
        ),
        (
            500,
            {
                "type": "server_error",
                "errors": [{"attr": None, "code": "oh_no", "detail": "Broken"}],
            },
            ArkindexServerError,
        ),
        # Errors that do not follow the standardized error structure fall back to an ErrorResponse
        (400, {}, ErrorResponse),
        (400, [{"type": "client_error"}], ErrorResponse),
        (400, [{"type": "client_error", "errors": []}], ErrorResponse),
        (400, {"type": "client_error", "errors": None}, ErrorResponse),
        (
            400,
            {
                "type": "cursed_error",
                "errors": [{"attr": None, "code": "invalid", "detail": "Wrong!"}],
            },
            ErrorResponse,
        ),
        (
            400,
            {
                "type": {"unexpected": "dict"},
                "errors": [{"attr": None, "code": "invalid", "detail": "Wrong!"}],
            },
            ErrorResponse,
        ),
        (
            400,
            {"errors": [{"attr": None, "code": "invalid", "detail": "Wrong!"}]},
            ErrorResponse,
        ),
        (
            400,
            {
                "type": "client_error",
                "errors": [{"attr": None, "code": "invalid", "detail": "Wrong!"}],
            },
            ErrorResponse,
        ),
        (
            400,
            {
                "type": "client_error",
                "errors": [
                    {
                        "attr": {"unexpected": "dict"},
                        "code": "invalid",
                        "detail": "Wrong!",
                    }
                ],
            },
            ErrorResponse,
        ),
        (
            400,
            {"type": "client_error", "errors": [{"attr": None, "detail": "Wrong!"}]},
            ErrorResponse,
        ),
        (
            400,
            {"type": "client_error", "errors": [{"attr": None, "code": "invalid"}]},
            ErrorResponse,
        ),
        (
            400,
            {
                "type": "client_error",
                "errors": [{"attr": None, "code": 42, "detail": "Wrong!"}],
            },
            ErrorResponse,
        ),
        (
            400,
            {
                "type": "client_error",
                "errors": [{"attr": None, "code": "invalid", "detail": ["Wrong!"]}],
            },
            ErrorResponse,
        ),
        (
            400,
            {
                "type": "client_error",
                "errors": [{"attr": [None], "code": "invalid", "detail": "Wrong!"}],
            },
            ErrorResponse,
        ),
    ],
)
def test_standardized_errors(responses, api_client, status_code, data, expected_class):
    responses.add(
        responses.POST,
        "https://dummy.test/api/v1/user/login/",
        status=status_code,
        json=data,
    )

    with pytest.raises(expected_class) as exc_info:
        api_client.request("Login", body={"email": "a", "password": "a"})

    assert exc_info.value.status_code == status_code
    reason = HTTPStatus(status_code).phrase
    assert exc_info.value.title == f"{status_code} {reason}"
    assert exc_info.value.content == data



from collections import namedtuple
from typing import Any, Literal, TypedDict

Position = namedtuple("Position", ["line_no", "column_no", "index"])


class ErrorMessage:
    def __init__(self, text, code, index=None, position=None):
        self.text = text
        self.code = code
        self.index = index
        self.position = position

    def __eq__(self, other) -> bool:
        return (
            self.text == other.text
            and self.code == other.code
            and self.index == other.index
            and self.position == other.position
        )

    def __repr__(self):
        return f"{self.__class__.__name__}({self.text!r}, code={self.code!r}, index={self.index!r}, position={self.position!r})"


class ErrorResponse(Exception):
    """
    Raised when a client request results in an error response being returned.
    """

    def __init__(self, title: str, status_code: int, content: Any) -> None:
        self.title = title
        self.status_code = status_code
        self.content = content

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.title!r}, status_code={self.status_code!r}, content={self.content!r})"

    def __str__(self) -> str:
        return repr(self.content)


class ClientError(Exception):
    """
    Raised when a client is unable to fulfil an API request.
    """

    def __init__(self, messages):
        self.messages = messages
        super().__init__(messages)

    def __repr__(self):
        return f"{self.__class__.__name__}(messages={self.messages!r})"

    def __str__(self):
        if len(self.messages) == 1 and not self.messages[0].index:
            return self.messages[0].text
        return str(self.messages)


class SchemaError(Exception):
    """
    Any error occurring during the acquisition and processing of the OpenAPI schema.
    """


class ArkindexErrorMessage(TypedDict):
    attr: list[str | int] | None
    """
    Optional path to an attribute that this error is related to.
    For example, `["elements", 1, "id"]` is equivalent to `body["elements"][1]["id"]`.

    This typically matches an attribute of the body of the API request, but it can also refer to the query parameters,
    HTTP headers, or to other attributes in the model that may be affected in any way by the API request.

    This attribute may be `None` for errors that do not apply only to one specific field.

    Due to limitations of the API framework, `["non_field_errors"]` should be treated the same as `None`.
    """

    code: str
    """
    A machine-readable error code. This may be used to identify specific errors automatically.
    """

    detail: str
    """
    A human-readable error message.

    This should only be used for display purposes and should not be parsed automatically.
    To detect specific errors, use `code` instead.
    """


class ArkindexErrorContent(TypedDict):
    """
    Content of an error response that follows the standard Arkindex error format.
    """

    type: Literal["server_error", "client_error", "validation_error"]
    errors: list[ArkindexErrorMessage]


class ArkindexConflictErrorContent(ArkindexErrorContent):
    existing_id: str | None


class ArkindexError(ErrorResponse):
    """
    An error response returned by the API that matches the standard Arkindex error format.
    """

    # Redefine __init__ explicitly so that the types are taken into account by type checkers
    # This all can be replaced with an ErrorResponse[T] generic class when we drop Python 3.10 support
    def __init__(self, title: str, status_code: int, content: ArkindexErrorContent):
        self.title = title
        self.status_code = status_code
        self.content = content


class ArkindexServerError(ArkindexError):
    """
    A server error response (HTTP 5xx) returned by the API that matches the standard Arkindex error format.

    This does not match every HTTP 5xx response, only those with this specific format,
    so this should not be used to handle transient errors and retry a request.

    To handle all HTTP 5xx responses, use `except ErrorResponse` and test the `status_code`.
    """


class ArkindexClientError(ArkindexError):
    """
    A client error response (HTTP 4xx) returned by the API that matches the standard Arkindex error format.
    """


class ArkindexValidationError(ArkindexClientError):
    """
    A validation error response returned by the API that matches the standard Arkindex error format.

    This is a subtype of client errors that only uses HTTP 400,
    where the `attr` of each error message is usually defined,
    and that usually refers to issues with the request's parameters.
    """

    def __init__(self, title: str, status_code: Literal[400], content: ArkindexErrorContent):
        self.title = title
        self.status_code = status_code
        self.content = content


class ArkindexConflictError(ArkindexClientError):
    """
    A conflict error response returned by the API that matches the standard Arkindex error format.

    This is a subtype of client errors that only uses HTTP 409 and may contain an `existing_id` attribute
    that refers to the ID of an existing resource that caused the conflict.
    """

    def __init__(
        self,
        title: str,
        status_code: Literal[409],
        content: ArkindexConflictErrorContent,
    ):
        self.title = title
        self.status_code = status_code
        self.content = content



import logging
from pathlib import Path

import requests
from requests import RequestException
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_chain,
    wait_fixed,
)

from arkindex import ArkindexClient, should_verify_cert
from arkindex.client.client import REQUEST_TIMEOUT
from arkindex.exceptions import ArkindexConflictError, ErrorResponse

logger = logging.getLogger(__name__)


@retry(
    reraise=True,
    retry=retry_if_exception_type(RequestException),
    stop=stop_after_attempt(3),
    wait=wait_fixed(5),
)
def upload_file(path: str | Path, url: str) -> None:
    with open(path, "rb") as f:
        # Directly use requests to upload the image, as the api_client.session.auth would conflict with the s3 API
        resp = requests.put(url, data=f, verify=should_verify_cert(url), timeout=REQUEST_TIMEOUT)
    resp.raise_for_status()


@retry(
    # Retry on any error, including HTTP 4xx because failed checks with handled errors can cause HTTP 400
    retry=retry_if_exception_type((RequestException, ErrorResponse)),
    wait=wait_chain(wait_fixed(10), wait_fixed(30)),
    stop=stop_after_attempt(3),
    reraise=True,
)
def check_image(api_client: ArkindexClient, image: dict, *, force: bool = False) -> dict:
    """
    Perform an image check on an existing image using `PartialUpdateImage`.

    :param api_client: An API client instance to use to make API requests.
    :type api_client: ArkindexClient
    :param image: A dict describing the image stored on Arkindex, as returned by `CreateImage` or `RetrieveImage`.
    :type image: dict
    :param force: Always run the image check, even if the image is already `checked`.
    :type force: bool
    :return: The updated image.
    :rtype: dict
    """
    if image["status"] == "checked" and not force:
        logger.debug(f"Skipping image check for image {image['id']} as it is already checked")
        return image

    logger.info(f"Checking image {image['id']}")
    image = api_client.request("PartialUpdateImage", id=image["id"], body={"status": "checked"})
    assert image["status"] == "checked", "Image was not marked as checked"
    return image


def create_image(
    api_client: ArkindexClient,
    path: str | Path,
    arkindex_path: str,
    *,
    force_check: bool = False,
) -> dict:
    """
    Upload an image from a local file into Arkindex. This handles already existing images and performs image checks.

    :param api_client: An API client instance to use to make API requests.
    :type api_client: ArkindexClient
    :param path: Path to a local file to use as the new image.
    :type path: Path
    :param arkindex_path: Destination path of the image within the local ImageServer.
        This is the `path` parameter of `CreateImage`.
    :type arkindex_path: str
    :param force_check: Call `PartialUpdateImage` to perform a new image check,
        even when the image is already marked as `checked`.
    :type force_check: bool
    :return: A dict describing the image stored on Arkindex, as returned by `CreateImage` or `RetrieveImage`.
    :rtype: dict
    """
    try:
        image = api_client.request("CreateImage", body={"path": arkindex_path})
    except ArkindexConflictError as e:
        # Handle errors on Arkindex ≥ 1.11.2
        if not e.content["existing_id"]:
            raise
        image = api_client.request("RetrieveImage", id=e.content["existing_id"])
    except ErrorResponse as e:
        # Handle errors on older versions
        if e.status_code != 400 or not isinstance(e.content, dict) or not e.content.get("id"):
            raise
        image = api_client.request("RetrieveImage", id=e.content["id"])

    if image["status"] != "checked":
        logger.info(f"Uploading image {image['id']}")
        assert image["s3_put_url"], "No S3 PUT URL was returned for this image."
        upload_file(path, image["s3_put_url"])

    return check_image(api_client, image, force=force_check)


def create_iiif_image(
    api_client: ArkindexClient,
    url: str,
    *,
    width: int | None = None,
    height: int | None = None,
) -> dict:
    """
    Import an IIIF image into Arkindex. This handles already existing images and performs image checks.

    :param api_client: An API client instance to use to make API requests.
    :type api_client: ArkindexClient
    :param url: The URL of the image's identifier.
       This should not contain any parameters like `/full/full/0/default.jpg`,
       and should not be the URL to the Image Information request (`/info.json`).
    :type url: str
    :param width: Width of the image in pixels.

       This is only allowed when the image is on a known ImageServer where external validation has been allowed,
       and when the user has the permission to trigger external validation. `height` must also be set.
    :type width: int | None
    :param height: Height of the image in pixels.

       This is only allowed when the image is on a known ImageServer where external validation has been allowed,
       and when the user has the permission to trigger external validation. `width` must also be set.
    :type height: int | None
    :return: A dict describing the image stored on Arkindex, as returned by `CreateIIIFURL` or `RetrieveImage`.
    :rtype: dict
    """
    assert not (width is None) ^ (height is None), "The width and height must either be both set or both unset."

    try:
        image = api_client.request(
            "CreateIIIFURL",
            body={"url": url, "width": width, "height": height},
        )
    except ArkindexConflictError as e:
        # Handle errors on Arkindex ≥ 1.11.2
        if not e.content["existing_id"]:
            raise
        image = api_client.request("RetrieveImage", id=e.content["existing_id"])
    except ErrorResponse as e:
        # Handle errors on older versions
        if e.status_code != 400 or not isinstance(e.content, dict) or not e.content.get("id"):
            raise
        image = api_client.request("RetrieveImage", id=e.content["id"])

    # Do not perform any check if the width and height were sent, because they are triggering an async check
    if width or height:
        return image
    return check_image(api_client, image)



import pytest

from arkindex.exceptions import ArkindexClientError, ArkindexConflictError, ErrorResponse
from arkindex.shortcuts import create_iiif_image, create_image

CONFLICT_EXCEPTIONS = [
    # Old format
    ErrorResponse(
        status_code=400, title="400 Bad Request", content={"path": ["An image already exists"], "id": "imageid"}
    ),
    # New format
    ArkindexConflictError(
        status_code=409,
        title="409 Conflict",
        content={
            "type": "client_error",
            "errors": [{"attr": ["path"], "code": "conflict", "detail": "An image already exists"}],
            "existing_id": "imageid",
        },
    ),
]


UNSUPPORTED_EXCEPTIONS = [
    # Errors without an ID
    ErrorResponse(status_code=400, title="400 Bad Request", content={"path": ["An image already exists"]}),
    ArkindexConflictError(
        status_code=409,
        title="409 Conflict",
        content={
            "type": "client_error",
            "errors": [{"attr": ["path"], "code": "conflict", "detail": "An image already exists"}],
            "existing_id": None,
        },
    ),
    # Non-conflict errors
    ErrorResponse(status_code=403, title="403 Forbidden", content={"detail": "Nope."}),
    ArkindexClientError(
        status_code=403,
        title="403 Forbidden",
        content={"type": "client_error", "errors": [{"attr": None, "code": "nope", "detail": "Nope."}]},
    ),
]


def test_create_image(mock_client, responses, tmp_path):
    file_path = tmp_path / "image.jpg"
    file_path.write_bytes(b"hello")
    mock_client.add_response(
        "CreateImage",
        body={"path": "image_path"},
        response={
            "id": "imageid",
            "status": "unchecked",
            "s3_put_url": "https://s3.dummy.test/puturl",
        },
    )
    responses.add(responses.PUT, "https://s3.dummy.test/puturl", body=b"hello", status=200)
    mock_client.add_response(
        "PartialUpdateImage", id="imageid", body={"status": "checked"}, response={"id": "imageid", "status": "checked"}
    )

    assert create_image(mock_client, file_path, "image_path") == {"id": "imageid", "status": "checked"}


@pytest.mark.parametrize("exception", CONFLICT_EXCEPTIONS)
def test_create_image_existing(mock_client, tmp_path, exception):
    file_path = tmp_path / "image.jpg"
    file_path.write_bytes(b"hello")
    mock_client.add_response(
        "CreateImage",
        body={"path": "image_path"},
        response=exception,
    )
    mock_client.add_response(
        "RetrieveImage",
        id="imageid",
        response={
            "id": "imageid",
            "status": "checked",
        },
    )
    assert create_image(mock_client, file_path, "image_path") == {"id": "imageid", "status": "checked"}


@pytest.mark.parametrize("exception", CONFLICT_EXCEPTIONS)
@pytest.mark.parametrize("image_status", ("unchecked", "error"))
def test_create_image_reupload(mock_client, responses, tmp_path, exception, image_status):
    file_path = tmp_path / "image.jpg"
    file_path.write_bytes(b"hello")
    mock_client.add_response(
        "CreateImage",
        body={"path": "image_path"},
        response=exception,
    )
    mock_client.add_response(
        "RetrieveImage",
        id="imageid",
        response={
            "id": "imageid",
            "status": image_status,
            "s3_put_url": "https://s3.dummy.test/puturl",
        },
    )
    responses.add(responses.PUT, "https://s3.dummy.test/puturl", body=b"hello", status=200)
    mock_client.add_response(
        "PartialUpdateImage", id="imageid", body={"status": "checked"}, response={"id": "imageid", "status": "checked"}
    )
    assert create_image(mock_client, file_path, "image_path") == {"id": "imageid", "status": "checked"}


@pytest.mark.parametrize("exception", CONFLICT_EXCEPTIONS)
def test_create_image_checked_force(mock_client, exception):
    """
    Calling create_image(…, force=True) on a checked existing image does not reupload the image, but it performs a check
    """
    mock_client.add_response(
        "CreateImage",
        body={"path": "image_path"},
        response=exception,
    )
    mock_client.add_response(
        "RetrieveImage",
        id="imageid",
        response={
            "id": "imageid",
            "status": "checked",
            "s3_put_url": "https://s3.dummy.test/puturl",
        },
    )
    mock_client.add_response(
        "PartialUpdateImage", id="imageid", body={"status": "checked"}, response={"id": "imageid", "status": "checked"}
    )
    assert create_image(mock_client, "/dev/null", "image_path", force_check=True) == {
        "id": "imageid",
        "status": "checked",
    }


@pytest.mark.parametrize("exception", UNSUPPORTED_EXCEPTIONS)
def test_create_image_error(mock_client, exception):
    mock_client.add_response(
        "CreateImage",
        body={"path": "image_path"},
        response=exception,
    )
    with pytest.raises(ErrorResponse) as ctx:
        create_image(mock_client, "/dev/null", "image_path")
    assert ctx.value == exception


@pytest.mark.parametrize(["width", "height"], [(1000, None), (None, 1000)])
def test_create_iiif_image_width_and_height(mock_client, width, height):
    with pytest.raises(AssertionError):
        create_iiif_image(mock_client, "http://dummy.test", width=width, height=height)


def test_create_iiif_image(mock_client):
    mock_client.add_response(
        "CreateIIIFURL",
        body={"url": "https://dummy.test", "width": None, "height": None},
        response={"id": "imageid", "status": "checked"},
    )
    assert create_iiif_image(mock_client, "https://dummy.test") == {"id": "imageid", "status": "checked"}


@pytest.mark.parametrize("exception", CONFLICT_EXCEPTIONS)
def test_create_iiif_image_exists(mock_client, exception):
    mock_client.add_response(
        "CreateIIIFURL",
        body={"url": "https://dummy.test", "width": None, "height": None},
        response=exception,
    )
    mock_client.add_response(
        "RetrieveImage",
        id="imageid",
        response={"id": "imageid", "status": "checked"},
    )
    assert create_iiif_image(mock_client, "https://dummy.test") == {"id": "imageid", "status": "checked"}


@pytest.mark.parametrize("exception", CONFLICT_EXCEPTIONS)
@pytest.mark.parametrize("image_status", ("unchecked", "error"))
def test_create_iiif_image_recheck(mock_client, exception, image_status):
    mock_client.add_response(
        "CreateIIIFURL",
        body={"url": "https://dummy.test", "width": None, "height": None},
        response=exception,
    )
    mock_client.add_response(
        "RetrieveImage",
        id="imageid",
        response={"id": "imageid", "status": image_status},
    )
    mock_client.add_response(
        "PartialUpdateImage",
        id="imageid",
        body={"status": "checked"},
        response={"id": "imageid", "status": "checked"},
    )
    assert create_iiif_image(mock_client, "https://dummy.test") == {"id": "imageid", "status": "checked"}


@pytest.mark.parametrize("exception", UNSUPPORTED_EXCEPTIONS)
def test_create_iiif_image_error(mock_client, exception):
    mock_client.add_response(
        "CreateIIIFURL",
        body={"url": "https://dummy.test", "width": None, "height": None},
        response=exception,
    )
    with pytest.raises(ErrorResponse) as ctx:
        create_iiif_image(mock_client, "https://dummy.test")
    assert ctx.value == exception


@pytest.mark.parametrize("exception", CONFLICT_EXCEPTIONS)
@pytest.mark.parametrize("image_status", ("unchecked", "error"))
def test_create_iiif_image_async_check(mock_client, exception, image_status):
    mock_client.add_response(
        "CreateIIIFURL",
        body={"url": "https://dummy.test", "width": 1000, "height": 2000},
        response=exception,
    )
    mock_client.add_response(
        "RetrieveImage",
        id="imageid",
        response={"id": "imageid", "status": image_status},
    )
    # PartialUpdateImage is not called, since the check will be done asynchronously
    assert create_iiif_image(mock_client, "https://dummy.test", width=1000, height=2000) == {
        "id": "imageid",
        "status": image_status,
    }
